/// Josh's 3D Math Library
/// A C++20 Library for 3D Math, Computer Graphics, and Scientific Computing.
/// Developed and Maintained by Josh O'Leary @ Redacted Software.
/// Special Thanks to William Tomasine II and Maxine Hayes.
/// (c) 2024 Redacted Software
/// This work is dedicated to the public domain.

/// @file J3ML.cpp
/// @edit 2024-07-05

#include <format>
#include <iomanip>
#include <strstream>
#include "J3ML/J3ML.hpp"

#include <sstream>


float PowUInt(float base, u32 exponent)
{
    // 'Fast Exponentiation': We interpret exponent in base two and calculate the power by
    // squaring and multiplying by base.

    // Find the highest bit that is set.
    u32 e = 0x80000000;
    while((exponent & e) == 0 && e > 0)
        e >>= 1;

    float val = 1.f;
    do
    {
        val *= val; // Shifts the exponent one place left
        val *= (exponent & e) != 0 ? base : 1.f; // Adds a 1 as the LSB of the exponent
        e >>= 1;
    } while(e > 0);

    return val;
}


namespace J3ML
{

    float Math::Functions::Radians(float degrees) { return degrees * (Pi/180.f); }

    float Math::Functions::Degrees(float radians) { return radians * (180.f/Pi); }



    Math::Rotation Math::operator ""_degrees(long double rads) { return {Functions::Radians((float)rads)}; }

    Math::Rotation Math::operator ""_deg(long double rads) { return {Functions::Radians((float)rads)}; }

    Math::Rotation Math::operator ""_radians(long double rads) { return {(float)rads}; }

    Math::Rotation Math::operator ""_rad(long double rads) { return {(float)rads}; }

    float Math::Functions::FastRSqrt(float x) {
        return 1.f / FastSqrt(x);
    }

    float Math::Functions::PingPongMod(float x, float mod) {
        x = Mod(x, mod * 2.f);
        return x >= mod ? (2.f * mod - x) : x;
    }

    float Math::Functions::Sqrt(float x) {
        return std::sqrt(x);
    }

    float Math::Functions::FastSqrt(float x) {
        // TODO: implement FastSqrt from MGL
        return std::sqrt(x);
    }

    float Math::Functions::RSqrt(float x) {
        return 1.f / Sqrt(x);
    }




    int SigFigsTable[] = {0,0,0,1,0,0,1,0,0,1};

    int DivBy[] = {1,1,1, 1000,1000,1000, 1000000, 1000000, 1000000, 1000000000, 1000000000,1000000000};

    std::vector<std::string> Suffixes = {
        "", "", "",
        "K", "K", "K",
        "M", "M", "M",
        "B", "B", "B",
        "T", "T", "T",
        "Q", "Q", "Q"
    };

    float Math::Functions::Round(float f, float decimalPlaces) {
        float mult = Pow(10, decimalPlaces);
        return Floor(f * mult + 0.5f) / mult;
    }

    float Math::Functions::Sign(float f) { return f >= 0.f ? 1.f : -1.f;}

    std::string Math::Functions::Truncate(float input) {

        std::stringstream ss;
        std::string str = "";

        if (input < 1000)
            ss << std::fixed << std::setprecision(0) << input;
        else {
            int figs = CeilInt(Log10(input)) - 1;
            auto suffix = Suffixes[figs];
            auto roundTo = SigFigsTable[figs];
            auto divBy = DivBy[figs];
            auto fractional = input / (float)divBy;

            // Increment roundTo for extra precision!!

            ss << std::fixed << std::setprecision(roundTo) << fractional << suffix;
        }



        str = ss.str();
        return str;

    }

    bool Math::Functions::Equal(float a, float b, float epsilon) {
        return Abs(a-b) <= epsilon;
    }

    bool Math::Functions::Equal(double a, double b, float epsilon) {
        return Abs(a - b) <= epsilon;
    }

    float Math::Functions::RelativeError(float a, float b) {
        if (a == b) return 0.f; // Handles the special case where approximation and real are both zero.
        return Abs((a-b) / Max(Abs(a), Abs(b)));
    }

    bool Math::Functions::EqualAbs(float a, float b, float epsilon) {
        // TODO: No different from Equal?
        return Abs(a - b) < epsilon;
    }

    float Math::Functions::RecipFast(float x) {
        // TODO: Implement SSE rcp instruction.
        return 1.f / x;
    }

    float Math::Functions::Lerp(float a, float b, float t) { return a + t * (b-a);}

    float Math::Functions::LerpMod(float a, float b, float mod, float t) {
        a = ModPos(a, mod);
        b = ModPos(b, mod);
        if (Abs(b - a) * 2.f <= mod)
            return Lerp(a, b, t);
        else {
            if (a < b)
                return ModPos(Lerp(a + mod, b, t), mod);
            else
                return ModPos(Lerp(a, b + mod, t), mod);
        }
    }

    float Math::Functions::InverseLerp(float a, float b, float x) {
        assert(Abs(b - a) > 1e-5f);
        return (x - a) / (b - a);
    }


    float Math::Functions::Step(float y, float x) {
        return (x >= y) ? 1.f : 0.f;
    }

    float Math::Functions::Ramp(float min, float max, float x) {

        return x <= min ? 0.f : (x >= max ? 1.f : (x - min) / (max - min));
    }

    float Math::Functions::Mod(float x, float mod) {
        return std::fmod(x, mod);
    }

    float Math::Functions::Mod(float x, int mod) {
        // TODO: Optimize?
        return std::fmod(x,(float)mod);
    }

    float Math::Functions::ModPos(float x, float mod) {
        float m = fmod(x, mod);
        return m >= 0.f ? m : (m + mod);
    }

    float Math::Functions::ModPos(float x, int mod) {
        return ModPos(x, (float)mod);
    }

    float Math::Functions::Frac(float x) {
        return x - Floor(x);
    }

    float Math::Functions::Recip(float x) {
        return 1.f / x;
    }



    Math::Rotation::Rotation() : valueInRadians(0) {}

    Math::Rotation::Rotation(float value) : valueInRadians(value) {}

    Math::Rotation Math::Rotation::operator+(const Math::Rotation &rhs) {
        return {valueInRadians + rhs.valueInRadians};
    }

    float Math::Interp::SmoothStart(float t) {
        assert(t >= 0.f && t <= 1.f);
        return t*t;
    }

    int Math::BitTwiddling::CountBitsSet(u32 value) {

    }

    namespace Math::Functions {
        float Sin(float x) {
#ifdef USE_LOOKUP_TABLES
#elif USE_SSE
#else
            return std::sin(x);
#endif
        }

        float Cos(float x) {
#ifdef USE_LOOKUP_TABLES
#elif USE_SSE
#else
            return std::cos(x);
#endif
        }

        float Tan(float x) { return std::tan(x); }

        void SinCos(float x, float &outSin, float &outCos) {
            outSin = Sin(x);
            outCos = Cos(x);
        }

        float Asin(float x) { return std::asin(x); }

        float Acos(float x) { return std::acos(x); }

        float Atan(float x) { return std::atan(x); }

        float Atan2(float y, float x) { return std::atan2(y, x); }

        float Sinh(float x) { return std::sinh(x); }

        float Cosh(float x) { return std::cosh(x); }

        float Tanh(float x) { return std::tanh(x); }

        bool IsPow2(u32 number) {
            return (number & (number - 1)) == 0;
        }

        bool IsPow2(u64 number) {
            return (number & (number - 1)) == 0;
        }

        float PowInt(float base, int exponent) {
            if (exponent == 0)
                return 1.f;
            else if (exponent < 0)
                return 1.f / PowUInt(base, (u32) -exponent);
            else return PowUInt(base, (u32) exponent);
        }

        float Pow(float base, float exponent) { return std::pow(base, exponent); }

        float Exp(float exp) { return std::exp(exp); }

        float Log(float base, float value) {
            return std::log(value) / std::log(base);
        }

        float Log2(float value) {
            return Log(2.f, value);
        }

        float Ln(float value) {
            return std::log(value);
        }

        float Log10(float value) {
            return Log(10, value);
        }

        float Ceil(float f) { return std::ceil(f); }
        int CeilInt(float f) { return (int)std::ceil(f);}

        float Floor(float x) { return std::floor(x); }

        int FloorInt(float x) { return (int)std::floor(x); }

        float Round(float x) { return Floor(x+0.5f); }

        int RoundInt(float x) { return (int)Round(x); }

    }
}
